@@ -43,7 +43,7 @@ struct menu_state {
 	struct wl_compositor *compositor;
 	struct wl_shm *shm;
 	struct wl_seat *seat;
-	struct wl_keyboard *keyboard;
+	struct wl_data_device_manager *data_device_manager;
 	struct zwlr_layer_shell_v1 *layer_shell;
 
 	struct wl_surface *surface;
@@ -53,6 +53,8 @@ struct menu_state {
 	struct xkb_keymap *xkb_keymap;
 	struct xkb_state *xkb_state;
 
+	struct wl_data_offer *offer;
+
 	struct pool_buffer buffers[2];
 	struct pool_buffer *current;
 
@@ -564,6 +566,36 @@ void keypress(struct menu_state *state, enum wl_keyboard_key_state key_state,
 			}
 			render_frame(state);
 			return;
+		case XKB_KEY_Y:
+			// Paste clipboard
+			if (!state->offer) {
+				return;
+			}
+
+			int fds[2];
+			if (pipe(fds) == -1) {
+				// Pipe failed
+				return;
+			}
+			wl_data_offer_receive(state->offer, "text/plain", fds[1]);
+			close(fds[1]);
+
+			wl_display_roundtrip(state->display);
+
+			while (true) {
+				char buf[1024];
+				ssize_t n = read(fds[0], buf, sizeof(buf));
+				if (n <= 0) {
+					break;
+				}
+				insert(state, buf, n);
+			}
+			close(fds[0]);
+
+			wl_data_offer_destroy(state->offer);
+			state->offer = NULL;
+			render_frame(state);
+			return;
 		case XKB_KEY_Left:
 		case XKB_KEY_KP_Left:
 			// Move to beginning of word
@@ -838,16 +870,31 @@ static void seat_capabilities(void *data, struct wl_seat *seat,
 		enum wl_seat_capability caps) {
 	struct menu_state *state = data;
 	if (caps & WL_SEAT_CAPABILITY_KEYBOARD) {
-		state->keyboard = wl_seat_get_keyboard(seat);
-		wl_keyboard_add_listener(state->keyboard, &keyboard_listener, state);
+		struct wl_keyboard *keyboard = wl_seat_get_keyboard(seat);
+		wl_keyboard_add_listener(keyboard, &keyboard_listener, state);
 	}
 }
 
-const struct wl_seat_listener seat_listener = {
+static const struct wl_seat_listener seat_listener = {
 	.capabilities = seat_capabilities,
 	.name = noop,
 };
 
+static void data_device_selection(void *data, struct wl_data_device *data_device,
+		struct wl_data_offer *offer) {
+	struct menu_state *state = data;
+	state->offer = offer;
+}
+
+static const struct wl_data_device_listener data_device_listener = {
+	.data_offer = noop,
+	.enter = noop,
+	.leave = noop,
+	.motion = noop,
+	.drop = noop,
+	.selection = data_device_selection,
+};
+
 static void handle_global(void *data, struct wl_registry *registry,
 		uint32_t name, const char *interface, uint32_t version) {
 	struct menu_state *state = data;
@@ -857,9 +904,11 @@ static void handle_global(void *data, struct wl_registry *registry,
 	} else if (strcmp(interface, wl_shm_interface.name) == 0) {
 		state->shm = wl_registry_bind(registry, name, &wl_shm_interface, 1);
 	} else if (strcmp(interface, wl_seat_interface.name) == 0) {
-		struct wl_seat *seat = wl_registry_bind(registry, name,
-				&wl_seat_interface, 4);
-		wl_seat_add_listener(seat, &seat_listener, state);
+		state->seat = wl_registry_bind(registry, name, &wl_seat_interface, 4);
+		wl_seat_add_listener(state->seat, &seat_listener, state);
+	} else if (strcmp(interface, wl_data_device_manager_interface.name) == 0) {
+		state->data_device_manager = wl_registry_bind(registry, name,
+				&wl_data_device_manager_interface, 3);
 	} else if (strcmp(interface, zwlr_layer_shell_v1_interface.name) == 0) {
 		state->layer_shell = wl_registry_bind(registry, name,
 				&zwlr_layer_shell_v1_interface, 1);
@@ -885,19 +934,19 @@ void insert(struct menu_state *state, const char *s, ssize_t n) {
 	}
 	memmove(state->text + state->cursor + n, state->text + state->cursor,
 			sizeof state->text - state->cursor - MAX(n, 0));
-	if (n > 0) {
+	if (n > 0 && s != NULL) {
 		memcpy(state->text + state->cursor, s, n);
 	}
 	state->cursor += n;
 	match(state);
 }
 
-char * fstrstr(struct menu_state *state, const char *s, const char *sub) {
+const char * fstrstr(struct menu_state *state, const char *s, const char *sub) {
 	size_t len;
 
 	for(len = strlen(sub); *s; s++)
 		if(!state->fstrncmp(s, sub, len))
-			return (char *)s;
+			return s;
 	return NULL;
 }
 
@@ -1016,8 +1065,15 @@ static void menu_init(struct menu_state *state) {
 	wl_registry_add_listener(registry, &registry_listener, state);
 	wl_display_roundtrip(state->display);
 	assert(state->compositor != NULL);
-	assert(state->layer_shell != NULL);
 	assert(state->shm != NULL);
+	assert(state->seat != NULL);
+	assert(state->data_device_manager != NULL);
+	assert(state->layer_shell != NULL);
+
+	// Get data device for seat
+	struct wl_data_device *data_device = wl_data_device_manager_get_data_device(
+			state->data_device_manager, state->seat);
+	wl_data_device_add_listener(data_device, &data_device_listener, state);
 
 	// Second roundtrip for xdg-output
 	wl_display_roundtrip(state->display);
@@ -1063,12 +1119,12 @@ bool parse_color(const char *color, uint32_t *result) {
 	if (color[0] == '#') {
 		++color;
 	}
-	int len = strlen(color);
+	size_t len = strlen(color);
 	if ((len != 6 && len != 8) || !isxdigit(color[0]) || !isxdigit(color[1])) {
 		return false;
 	}
 	char *ptr;
-	uint32_t parsed = strtoul(color, &ptr, 16);
+	uint32_t parsed = (uint32_t)strtoul(color, &ptr, 16);
 	if (*ptr != '\0') {
 		return false;
 	}
@@ -1172,7 +1228,7 @@ int main(int argc, char **argv) {
 		{ wl_display_get_fd(state.display), POLLIN },
 		{ state.repeat_timer, POLLIN },
 	};
-	const int nfds = sizeof(fds) / sizeof(*fds);
+	const size_t nfds = sizeof(fds) / sizeof(*fds);
 
 	while (state.run) {
 		errno = 0;
